import dateutil.parser
from .api import Api

class Vehicle(object):

    def __init__(self, vehicle_id, access_token, unit_system='metric'):
        """ Initializes a new Vehicle to use for making requests to the Smartcar API.

        Args:
            vehicle_id (str): the vehicle's unique identifier
            access_token (str): a valid access token
            unit_system (str, optional): the unit system to use for vehicle data.
                Defaults to metric.

        """
        self.vehicle_id = vehicle_id
        self.access_token = access_token
        self.api = Api(access_token, vehicle_id)
        self.api.set_unit('metric' if unit_system == 'metric' else 'imperial')

    def set_unit(self, unit):
        """ Update the unit system to use in requests to the Smartcar API.

        Args:
            unit (str): the unit system to use (metric/imperial)

        """
        if unit not in ('metric','imperial'):
            raise ValueError("unit must be either metric or imperial")
        else:
            self.api.set_unit(unit)

    def info(self):
        """ GET Vehicle.info

        Returns:
            dict: vehicle's info

        """
        response = self.api.get('')

        return response.json()

    def vin(self):
        """ GET Vehicle.vin

        Returns:
            str: vehicle's vin
        """
        response = self.api.get('vin')

        return response.json()['vin']

    def permissions(self):
        """ GET Vehicle.permissions

        Returns:
            list: vehicle's permissions
        """
        response = self.api.permissions()

        return response.json()['permissions']

    def disconnect(self):
        """ Disconnect this vehicle from the connected application.

        Note: Calling this method will invalidate your access token and you will
        have to have the user reauthorize the vehicle to your application if you
        wish to make requests to it

        """
        self.api.disconnect()

    def odometer(self):
        """ GET Vehicle.odometer

        Returns:
            dict: vehicle's odometer

        """
        response = self.api.get('odometer')

        return {
            'data': response.json(),
            'unit_system': self.api.unit,
            'age': dateutil.parser.parse(response.headers['sc-data-age']),
        }

    def location(self):
        """ GET Vehicle.location

        Returns:
            dict: vehicle's location

        """
        response = self.api.get('location')

        return {
            'data': response.json(),
            'age': dateutil.parser.parse(response.headers['sc-data-age']),
        }

    def unlock(self):
        """ POST Vehicle.unlock

        """
        self.api.action('security', 'UNLOCK')

    def lock(self):
        """ POST Vehicle.lock

        """
        self.api.action('security', 'LOCK')



API_VERSION = '1.0'
API_URL = 'https://api.smartcar.com/v{}'.format(API_VERSION)
AUTH_URL = 'https://auth.smartcar.com/oauth/token'
CONNECT_URL = 'https://connect.smartcar.com'
UNIT_HEADER = 'sc-unit-system'



from . import api, const, requester, vehicle
import time
from datetime import datetime, timedelta
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode

def set_expiration(access):
    expire_date = datetime.utcnow() + timedelta(seconds=access["expires_in"])
    refresh_expire_date = datetime.utcnow() + timedelta(days=60)
    access['expiration'] = expire_date
    access['refresh_expiration'] = refresh_expire_date
    return access

def is_expired(expiration):
    """ Check if an expiration is expired

    Args:
        expiration (datetime): expiration datetime

    Returns:
        bool: true if expired
    """
    return datetime.utcnow() > expiration

def get_vehicle_ids(access_token, limit=10, offset=0):
    """ Get a list of the user's vehicle ids

    Args:
        access_token (str): A valid access token from a previously retrieved
            access object
        limit (integer, optional): The number of vehicle ids to return
        offset (integer, optional): The index to start the vehicle list at

    Returns:
        dict: response containing the list of vehicle ids and paging information

    """
    return api.Api(access_token).vehicles(limit=limit, offset=offset).json()

def get_user_id(access_token):
    """ Retrieve the userId associated with the access_token

    Args:
        access_token (str): Smartcar access token

    Returns:
        str: userId

    """
    return api.Api(access_token).user().json()['id']

class AuthClient(object):

    def __init__(self, client_id, client_secret, redirect_uri, scope=None, development=False):
        """ A client for accessing the Smartcar API

        Args:
            client_id (str): The application id, provided in the application
                dashboard
            client_secret (str): The application secret, provided in the
                application dashboard
            redirect_uri (str): The URL to redirect to after the user accepts
                or declines the application's permissions. This URL must also be
                present in the Redirect URIs field in the application dashboard
            scope (bool, optional): A list of permissions requested by the application
            development (bool, optional): If True, deplays the Mock OEM for testing.
                Defaults to False

        """
        self.client_id = client_id
        self.client_secret = client_secret
        self.auth=(client_id, client_secret)
        self.redirect_uri = redirect_uri
        self.scope = scope
        self.development = development

    def get_auth_url(self, force=False, state=None):
        """ Generate an OAuth authentication URL

        Args:
            force (bool, optional): Set to True in order to force the approval
                dialog shown to the user. Defaults to False.
            state (bool, optional): A random string that will be passed back on
                redirect, this allows protection against cross-site forgery
                requests. Defaults to None.

        Returns:
            str: authorization url

        """
        base_url = const.CONNECT_URL

        approval_prompt = 'force' if force else 'auto'
        query = {
            'response_type': 'code',
            'client_id': self.client_id,
            'redirect_uri': self.redirect_uri,
            'approval_prompt': approval_prompt,
            'development': self.development
        }

        if self.scope:
            query['scope'] = ' '.join(self.scope)

        if state:
            query['state'] = state

        return base_url + '/oauth/authorize?' + urlencode(query)

    def exchange_code(self, code):
        """ Exchange an authentication code for an access dictionary

        Args:
            code (str): A valid authorization code

        Returns:
            dict: dict containing the access and refresh token

        """
        method = 'POST'
        url = const.AUTH_URL
        data = {
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': self.redirect_uri,
        }
        response = requester.call(method, url, data=data, auth=self.auth).json()
        return set_expiration(response)


    def exchange_refresh_token(self, refresh_token):
        """ Exchange a refresh token for a new access dictionary

        Args:
            refresh_token (str): A valid refresh token from a previously retrieved
                access object

        Returns:
            dict: dict containing access and refresh token

        """
        method = 'POST'
        url = const.AUTH_URL
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }
        response = requester.call(method, url, data=data, auth=self.auth).json()
        return set_expiration(response)



import requests
from . import exceptions as E
from . import __version__

def call(method, url, **kwargs):
    """ Attachs the kwargs into the headers, sends the request to the Smartcar API
        and handles all error cases

    Args:
        method (str): HTTP method
        url (str): url of the request
        **kwargs: parameters for the request

    Returns:
        dict: response from the request to the Smartcar API

    """
    if not 'headers' in kwargs:
        kwargs['headers'] = {}
    kwargs['headers']['User-Agent'] = 'smartcar-python-sdk:{}'.format(__version__)

    response = requests.request(method, url, **kwargs)
    code = response.status_code
    if response.ok:
        return response
    elif code == 400:
        raise E.ValidationException(response)
    elif code == 401:
        raise E.AuthenticationException(response)
    elif code == 403:
        raise E.PermissionException(response)
    elif code == 404:
        raise E.ResourceNotFoundException(response)
    elif code == 409:
        raise E.StateException(response)
    elif code == 429:
        raise E.RateLimitingException(response)
    elif code == 430:
        raise E.MonthlyLimitExceeded(response)
    elif code == 500:
        raise E.ServerException(response)
    elif code == 501:
        raise E.NotCapableException(response)
    elif code == 504:
        raise E.GatewayTimeoutException(response)
    else:
        response.raise_for_status()



from . import const, requester

class Api(object):

    def __init__(self, access_token, vehicle_id=None):
        """ Initialize a new Api object to make directly make requests to Smartcar.

        Args:
            access_token (str): Smartcar access token
            vehicle_id (str, optional): id of the vehicle. Defaults to None.
        """
        self.access_token = access_token
        self.vehicle_id = vehicle_id
        self.auth = {
            'Authorization': 'Bearer {}'.format(access_token)
        }
        self.unit = 'metric'

    def set_unit(self, unit):
        """ Update the unit system to use in requests to the Smartcar API.

        Args:
            unit (str): the unit system to use (metric/imperial)

        """
        self.unit = unit

    def _format(self, endpoint):
        """ Generates the formated URL

        Args:
            endpoint (str): the Smartcar endpoint of interest

        Returns:
            str: formatted url

        """
        return '{}/{}/{}'.format(const.API_URL, self.vehicle_id, endpoint)

    def action(self, endpoint, action, **kwargs):
        """ Sends POST requests to Smartcar API

        Args:
            endpoint (str): the Smartcar endpoint of interest
            action (str): action to be taken
            **kwargs: information to put into the body of the request

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = self._format(endpoint)
        headers = self.auth
        headers[const.UNIT_HEADER] = self.unit
        json = { 'action': action }
        for k,v in kwargs.items():
            if v:
                json[k] = v

        return requester.call('POST', url, json=json, headers=self.auth)

    def get(self, endpoint):
        """ Sends GET requests to Smartcar API

        Args:
            endpoint (str): the Smartcar endpoint of interest

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = self._format(endpoint)
        headers = self.auth
        headers[const.UNIT_HEADER] = self.unit
        return requester.call('GET', url, headers=headers)

    def permissions(self, **params):
        """ Sends a request to /permissions

        Args:
            **params: parameters for the request

        Returns:
            Reponse: response from the request to the Smartcar API

        """
        url = self._format('permissions')
        return requester.call('GET', url, headers=self.auth, params=params)

    def disconnect(self):
        """ Sends a request to /application

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = self._format('application')
        return requester.call('DELETE', url, headers=self.auth)

    def vehicles(self, **params):
        """ Sends a request to /vehicles

        Args:
            **params: parameters for the request

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = '{}/{}'.format(const.API_URL, 'vehicles')
        return requester.call('GET', url, headers=self.auth, params=params)

    def user(self, **params):
        """ Sends a request to /user

        Args:
            **params: parameters for the request

        Returns:
            Response: response from the request to the Smartcar API

        """
        url = '{}/{}'.format(const.API_URL, 'user')
        return requester.call('GET', url, headers=self.auth, params=params)



class SmartcarException(Exception):
    def __init__(self, response):
        json = response.json()
        if 'message' in json:
            self.message = json['message']
        elif 'error_description' in json:
            self.message = json['error_description']
        else:
            self.message = 'Unknown error'

    def __str__(self):
        return self.message

class ValidationException(SmartcarException):
    pass
class AuthenticationException(SmartcarException):
    pass
class PermissionException(SmartcarException):
    pass
class ResourceNotFoundException(SmartcarException):
    pass
class StateException(SmartcarException):
    pass
class RateLimitingException(SmartcarException):
    pass
class MonthlyLimitExceeded(SmartcarException):
    pass
class ServerException(SmartcarException):
    pass
class NotCapableException(SmartcarException):
    pass
class GatewayTimeoutException(SmartcarException):
    def __init__(self, response):
        self.message = response.text

    def __str__(self):
        return self.message



from setuptools import setup
import re
with open('smartcar/__init__.py','r') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        f.read(), re.MULTILINE).group(1)
setup(
    name='smartcar',
    version=version,
    description='Smartcar Python SDK',
    author='Smartcar',
    author_email='hello@smartcar.com',
    packages=['smartcar'],
    url='https://github.com/smartcar/python-sdk',
    license='MIT',
    install_requires=[
        'python-dateutil',
        'requests',
    ],
    extras_require={
        'dev': [
            'coverage',
            'nose',
            'mock',
            'responses',
        ]
    }
)



__version__ = '1.0.0'
from .const import (API_VERSION, API_URL, AUTH_URL, UNIT_HEADER)
from .smartcar import (AuthClient, is_expired, get_user_id, get_vehicle_ids)
from .vehicle import Vehicle
from .exceptions import (
    SmartcarException, ValidationException, AuthenticationException,
    PermissionException, ResourceNotFoundException, StateException,
    RateLimitingException, MonthlyLimitExceeded, ServerException,
    NotCapableException, GatewayTimeoutException
)
